//========================================================================================
// Mer.hh
// ---------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
// $Id: //CBT-depot/CBT++/include/Mer.hh#28 $
//----------------------------------------------------------------------------------------
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//

#ifndef _MER
#define _MER

#include "CBT.hh"
#include "Sequence3Bit.hh"

#define ACode 0
#define CCode 1
#define GCode 2
#define TCode 3

#define BITSPERBASE (2UL)
#define BASEMASK ((1UL << BITSPERBASE) - 1)

//----------------------------------------------------------------------------------------
// * Mer
//----------------------------------------------------------------------------------------
// Class to represent an n-mer up to 32bp with a 2-bit/bp encoding. This class is primarialy
// used for hashing n-mers where keeping the total space to a minimum is important.  Since 
// most systems we use have 32 bit ints, Mer uses two ints to represent a 32bp sequence. 
//
//      Hex         Dec            Binary   Encoding Bits
// a = 61 (hex)    97 (dec)      0110 0001   0110 0001   a -> 0
// c = 63 (hex)    99 (dec)      0110 0011   0110 0011   c -> 1
// g = 67 (hex)    103 (dec)     0110 0111   0110 0111   g -> 2
// t = 74 (hex)    116 (dec)     0111 0100   0111 0100   t -> 3

// A = 41 (hex)    65 (dec)      0100 0001   0100 0001   A -> 0
// C = 43 (hex)    67 (dec)      0100 0011   0100 0011   C -> 1
// G = 47 (hex)    71 (dec)      0100 0111   0100 0111   G -> 2 
// T = 54 (hex)    84 (dec)      0101 0100   0101 0100   T -> 3 
class Mer{
 public: 
  
  uint64  mData;
  
  // Mainly for space reasons, only allow a global definition of 
  // the mersize. 
  static int BASESUNUSED; 
  static int BITSUNUSED;
  static int mMerSize;
  static int mThinFactor;
  static int mThinSlice;
  
  Mer(){mData = 0;};
  Mer(Sequence &S,int i);
  Mer(Sequence &Seq,int merStart,bool &bContainsIllegal);

  Mer(Sequence3Bit &Seq,int merStart,bool &bContainsIllegal);
 
  void set(const char* s);
  
  // ShiftInBits
  inline void ShiftInBits(char b)
  {
    mData <<= (BITSPERBASE + BITSUNUSED);  
    mData >>= BITSUNUSED; 
    mData |= b;
  }

  // ASCII ONLY TEST.  But Fast!
  // n = 6e (hex)    110 (dec)     0110 1110   0110 (111)0   n -> 7
  // x = 78 (hex)    120 (dec)     0111 1000   0111 (100)0   x -> 4
  // N = 4e (hex)    78 (dec)      0100 1110   0100 (111)0   N -> 7
  // X = 58 (hex)    88 (dec)      0101 1000   0101 (100)0   X -> 4  
  // All the encodings of N or X have bit3 set.  So and with 8.  If it is non-zero 
  // it contains an N or X. 
  static inline bool isNX(char asciiBase){
    if (asciiBase & 0x0008) return(true);
    else return(false);
  }

  void InitWithFirstCleanMer(Sequence &S,int &merStart);
  void ShiftInClean(Sequence &S,int &i);
  void ShiftInBase(char b);

  
  static void setMerSize(int merSize){
    BASESUNUSED = 32 - merSize; 
    BITSUNUSED=(BASESUNUSED*BITSPERBASE);
    mMerSize = merSize;
  }

  static void setThinFactor(int thinFactor){
    mThinFactor = thinFactor;
  }

  static void setThinSlice(int thinSlice){
    mThinSlice = thinSlice;
  }

  static int getMerSize(){return(mMerSize);}
  static int getThinFactor(){return(mThinFactor);}
  static int getThinSlice(){return(mThinSlice);}

  
  //--------------------------------------
  // * Reverse
  //--------------------------------------
  // Reverse complement a word.  
  // 
  //inline void Reverse() {
  void Reverse() {
    register uint32 i;
    register uint64 rev = ~0;
    register uint64 mask = rev;
    register int msize = mMerSize;
   
    // The word mData is treated as a string of 2-bit bases
    // (A == 01, C == 01, G == 10, T == 11, 
    // So long as we reverse the string basewise,
    // we just need to bit-complement to get the final answer.  
    for (i = 1; i <= msize; i++) {
      //  Shift result left and add in last base of input; then advance input
      rev = (rev << BITSPERBASE) + (mData & BASEMASK);
      mData >>= BITSPERBASE;
    }

    // Unfortunately, if we just take complement of this, 
    // it will set all the unused bits to 1.  This is fine so far
    // things like ToString() are concerned, but MerHash simply looks
    // at these as numbers, nothing else.  So, for MerHash 
    // ffffffffffcab06c  is not the same mer as 00000000000ab06c
    // even though, as 11 mers, these are the same.    
   
    mData = ~rev;
    // So, need to explicitly set unused bits to zero... 
    mask = mask >> BITSUNUSED;
    mData = mData & mask;
  }

  string ToString() const;

  // Sometimes we already have Mer and rcMer in hand 
  // so pass it in to save computation. 
  //inline bool isThinned(Mer &rcMer){
  bool isThinned(Mer &rcMer){
    // From Paul's code:
    //if (w % HashSlicing != HashSlice) {
    //  return;
    //}
        // The canonical version of a mer is the lesser of the 
    // fwd or the rev, so compute that to evaluate thin on.
    Mer tempMer;
    tempMer = (mData < rcMer.mData) ? (*this) : rcMer;    
    return((tempMer.mData % mThinFactor) != mThinSlice);    
  }

  bool isThinned(){
    //inline bool isThinned(){
    // From Paul's code:
    //if (w % HashSlicing != HashSlice) {
    //  return;
    //}
   Mer rcMer = (*this);
   rcMer.Reverse();
   Mer tempMer;
   tempMer = (mData < rcMer.mData) ? (*this) : rcMer;    

   //   cout<<"\t\t mData="<<mData<<endl;
   //cout<<"\t\t rc.Data="<<rcMer.mData<<endl;
   //cout<<"\t\t tempMer.mData="<<tempMer.mData<<endl;
   //cout<<"\t\t tempMer.mData % "<<mThinFactor<<" = "<<mThinSlice<<endl;

   return((tempMer.mData % mThinFactor) != mThinSlice);    
  }


  // To save time (?) can pas in factor and slice as you call 
  // this.  Alternative is to call set methods or to pass it in
  // at the time you create the Mer. 
  //  inline bool isThinned(uint32 factor,uint32 slice){
  bool isThinned(uint32 factor,uint32 slice){
    // From Paul's code:
    //if (w % HashSlicing != HashSlice) {
    //  return;
    //}
    Mer rcMer = (*this);
    rcMer.Reverse();
    Mer tempMer;
    tempMer = (mData < rcMer.mData) ? (*this) : rcMer;    
    return((tempMer.mData % factor) != slice);    
  }

  
  // *** Operators 
  bool operator==(const Mer &m) const {
    return(m.mData == mData);
  }

  // *** Operators 
  bool operator!=(const Mer &m) const {
    return(m.mData != mData);
  }

  Mer(const char *m){
    set(m);
  }

  Mer(string &m){
    set(m.c_str());
  }
  /***** Not yet implemented. *******/

  void ToSequence(Sequence &Seq) const {cerr<<"Mer::ToSequence NYI"<<endl;exit(1);}
  /**********************************/

};


#endif


