//========================================================================================
// Mer2SequenceLocationMap.hh
// ---------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
// $Id: //CBT-depot/CBT++/include/Mer2SequenceLocationMap.hh#7 $
//----------------------------------------------------------------------------------------
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//

#ifndef _MER2SEQUENCELOCATIONMAP
#define _MER2SEQUENCELOCATIONMAP

#include "CBT.hh"
#include "Mer.hh"
#include "MerHashF.hh"
#include "Mer2IntMap.hh"
#include "MerRepeatModel.hh"

#include "SequenceLocation.hh"
#include "SearchResult.hh"

// KJD: Why is this here? 
//#define SEQ_BUFFER_SIZE2 500000
//extern char gseqbuf2[SEQ_BUFFER_SIZE2];

//----------------------------------------------------------------------------------------
// * SequenceLocationVecT
//----------------------------------------------------------------------------------------
// A vector of sequence locations.  The sequence location is space optimized
// for particular applications and these implementations are found in SequenceLocation.hh  
// Example applications include:
//
// Source is collection of reads --> Location is source idx + small bit base idx. 
// Source is single big source   --> Location is a large big base idx. 
//
// This will be the data part of a Mer2SourceLocationMap 
// 
template<class Loc>
class SequenceLocationVecT : public vector<Loc>{
public:
};

//----------------------------------------------------------------------------------------
// * Mer2SequenceLocationMap
//----------------------------------------------------------------------------------------
// The instantiated version of Mer2SequenceLocationMapT, with MerHashF as the hash function and 
// HitLocationVec as the sattelite information.  
//
// This instantiated version also keeps a local copy of the sequences used to create the 
// hash.  For space efficiency, this local copy is a 3-bit encoding of the sequence 
// provided by the class Sequence3Bit (3 bit because we still want to know Ns and Xs). 
// 
template<class H,class S,class ModelHashF = MerSymmetricHashF>
class Mer2SequenceLocationMap : public hash_map<Mer,SequenceLocationVecT<S> ,H,H>{
 public:
  

  int    mMaxMerLocations;  // Hard limit on locations stored. 

  // Statistics....
  long   mTotalMersInHash;

  // When a source hash is built, need to save source names and sizes 
  // so they can be referenced later.  The sourceIdx is saved in the SequenceLocation
  // to refer back to this longer name. 
  vector<string> mSourceNames;
  vector<int>    mSourceSizes;

  Mer2SequenceLocationMap(){
    mMaxMerLocations = 999999;
  }

  void Add(istream &fin){
    MerRepeatModel<ModelHashF> EmptyModel;
    Add(fin,EmptyModel);
  }
  void Add(istream &fin,MerRepeatModel<ModelHashF> &R);

  void AddMerLocation(Mer &TempMer, int SeqIdx, int BaseIdx, MerRepeatModel<ModelHashF> &R);

  void AddMerLocation(Mer &TempMer,int SeqIdx,int BaseIdx){
    MerRepeatModel<ModelHashF> EmptyModel;
    AddMerLocation(TempMer,SeqIdx,BaseIdx,EmptyModel);
  }

  void FindMerLocations(Mer &M,SearchResultVec &Results);    
};



//------------------------------------
// * AddMerLocation
//------------------------------------
// Adds TempMer to hash if it is not already there.  If it is already there, tacks on an 
// entry for SeqIdx,BaseIdx to the MerLocationVec for that Mer. 
//
//template<class ModelHashF, 
template<class H,class MerLoc,class ModelHashF>
void 
Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::AddMerLocation(Mer &TempMer,int SeqIdx,int BaseIdx,
						  MerRepeatModel<ModelHashF> &RepeatModel){

  //cout<<"ADDMERLOC: "<< SeqIdx<<"\t"<<BaseIdx<<"\t"<<TempMer.ToString()<<endl; 

  // Skip thinned mers...
  if (TempMer.isThinned()) return;

  // Skip mers which are too repetitive....
  // An empty repeat model will always return true. 
  if (!RepeatModel.isAcceptableFrequency(TempMer)) return;

  typename Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::iterator it;
  it = (*this).find(TempMer);

  MerLoc MMerLocation;
  MMerLocation.setSourceIdx(SeqIdx);
  MMerLocation.setSourceStart(BaseIdx);
  if (it != (*this).end()){    
    // If it's already there, then just add another info to the list.
    // There is a hard limit to minimize effects of extreme outliers. 
    int merLocs = (*it).second.size();
    if (merLocs < mMaxMerLocations){
      (*it).second.resize(merLocs+1);
      (*it).second[merLocs] = MMerLocation;
      mTotalMersInHash++;  // Used to compute a global average. 
    }
  }else{
    // Otherwise, add a MerLocations structure to the list
    SequenceLocationVecT<MerLoc> MMerLocations;
    MMerLocations.push_back(MMerLocation);
    (*this)[TempMer] = MMerLocations;
  }
}

//------------------------------------
// * Add
//------------------------------------
// Adds TempMer to hash if it is not already there.  If it is already there, tacks on an 
// entry for SeqIdx,BaseIdx to the MerLocationVec for that Mer. 
//

// Alternative logic...
 
/* 
do{
  char inchar = readbuf[readidx++];
  if (inchar == '>'){
    handle new read cases.
  }else if (inchar == '\0'){
    handle EOL condition. 
  }else{
    baseIdx++; 
    if (!Mer::isNX(inchar)){
      TempMer.ShiftInBase(inchar); 
      AddMerLocation(TempMer,seqIdx,baseIdx,RM); 
    }
  }
}

KJD The logic above is still slightly flawed, as is that below. 
The logic as given will introduce false Mers around Ns.  It doesn't
properly skip over the Mer.  Once I have finished making a similar 
mod to BangEngine, I'll propagate the changes here as well.  

*/

template<class H,class MerLoc,class ModelHashF>
void
Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::Add(istream &fin,MerRepeatModel<ModelHashF> &RM)
{
  Mer TempMer;
  long seqIdx = -1;
  long baseIdx = 0;
  int mersize = Mer::mMerSize;

  do{
    fin.getline(readbuf,READ_BUFFER_SIZE);
    int readidx = 0;
    bool bEOL = false;
    do{
      char inchar = readbuf[readidx++];
      switch(inchar){
	case '>':
	  // Store the read name, the last read size (unless this is the first)
	  // increment seqIdx and reset baseIdx
	  mSourceNames.push_back(readbuf+1);
	  //	  if (seqIdx != 0){
	  if (seqIdx != -1){
	    // Note that the last mer we will be able to add 
	    // will be Mer::mMerSize -1 bases from the end of the actual 
	    // sequence, so we need to add this to get the true size. 
	    //   cout<<"seqIdx="<<seqIdx<<"\tmSourceSizes.size="<<mSourceSizes.size()<<endl;
	    mSourceSizes.push_back(baseIdx+Mer::mMerSize);
	  }
	  seqIdx++;

	  //cout<<"Added "<<readbuf<<"\tmSourceNames.size="<<mSourceNames.size()<<"\tseqIdx="<<seqIdx<<endl;

	  baseIdx = 0;	  
	  // We're done with that buffer's worth, so grab a new one...
	  fin.getline(readbuf,READ_BUFFER_SIZE);
	  // Load up TempMer with the first k-mer worth of sequence and add it in. 
	  TempMer.set(readbuf);	  
	  readidx = TempMer.mMerSize;
	  AddMerLocation(TempMer,seqIdx,baseIdx,RM);
	  break;
	case 'A':
	case 'a':
	  baseIdx++;
	  TempMer.ShiftInBits(ACode); 
	  AddMerLocation(TempMer,seqIdx,baseIdx,RM);
	  break;
	case 'C':
	case 'c':
	  baseIdx++;
	  TempMer.ShiftInBits(CCode); 
	  AddMerLocation(TempMer,seqIdx,baseIdx,RM);
	  break;
	case 'G':
	case 'g':
	  baseIdx++;
	  TempMer.ShiftInBits(GCode); 
	  AddMerLocation(TempMer,seqIdx,baseIdx,RM);
	  break;
	case 'T':
	case 't':
	  baseIdx++;
	  TempMer.ShiftInBits(TCode); 
	  AddMerLocation(TempMer,seqIdx,baseIdx,RM);
	  break;

	case 'N':
	case 'n':
	case 'x':
	case 'X':
	  baseIdx++; // There is no mer to shift in, but need to count base. 
	  break;
	case '\0':
	  readidx = 0;
	  bEOL = true;
	  break;
	default:
	  break;	  
      }
    }while(!bEOL);
  }while(!fin.eof());
  // mMerSize is added because last mer starts before the actual end.
  //cout<<"Last seqIdx="<<seqIdx<<"\tmSourceSizes.size="<<mSourceSizes.size()<<endl; 
  mSourceSizes.push_back(baseIdx+Mer::mMerSize);  
}


//--------------------------------------
// * FindMerLocations
//--------------------------------------
// Finds all the locations of the mer specified by merStart in s. These locations are added 
// to the SearchResultVecVec.  
// 
template<class H,class MerLoc,class ModelHashF>
void
Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::FindMerLocations(Mer &tempMer,
					    SearchResultVec &Results)
{
  typename Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::iterator it;
  it = (*this).find(tempMer);      

  int count = 0;  
  // Concatenate the new results onto the existing Results vector.
  if (it != Mer2SequenceLocationMap<H,MerLoc,ModelHashF>::end()) {
    SequenceLocationVecT<MerLoc> &foundLocations = (*it).second;
    register int lsize = foundLocations.size();

    for(register int i = 0;i< lsize;i++){
      MerLoc &M = foundLocations[i];
      long sourceend = M.sourceStart()+Mer::mMerSize-1;

      // BUG test
      if (M.sourceIdx() >= mSourceNames.size()){
	cout<<"BUG mSourceNames.size="<<mSourceNames.size()<<endl;
	for(int b = 0;b < mSourceNames.size();b++) cout<<mSourceNames[b]<<endl;
	cout<<"BUG M.sourceIdx()="<<M.sourceIdx()<<endl;
      }

      string sourceName = mSourceNames[M.sourceIdx()];
      Results.add(sourceName,mSourceSizes[(int)M.sourceIdx()],M.sourceStart(),sourceend);		 
    }
  }
}

#endif 



//
// If I made MerRepeatModel a member of this class, would need to get a 
// copy, but it's too large to copy. 
//
// If I saved a pointer to the model, set with a setModel() method, then 
// the class would be vulnerable to someone deleting the model leaving us 
// with a bad pointer.  This could be at least partially addressed by 
// moving the model to the topmost level of an application, and passing the 
// pointer into the various classes that need it.  As a top level method, 
// the calling program would be responsible for keeping the class alive.  
// 
//
// RepeatModel R;
// BangEngine B(&R){
//     Mer2SeqLocationMap M(pR);
// }
// 


//
// If I saved the model here
