//========================================================================================
// MerHashF.hh
// ---------------------------------------------------------------------------------------
// Mer hash functors suitable for stl hash containers. 
// 
//----------------------------------------------------------------------------------------
// $Id: //CBT-depot/CBT++/include/MerHashF.hh#8 $
//----------------------------------------------------------------------------------------
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//

#ifndef _MERHASHF
#define _MERHASHF

#include "CBT.hh"
#include "Mer.hh"

//------------------------------------------
// * MerHashF
//------------------------------------------
// Computes the hash value of a mer.   This simple hash function computes the 
// XOR of bits of upper 32-mer with the bits of the lower 32-mer.  The XOR truth table has 
// 1/2 0s and 1/2 1s, so to 1st order (i.e. if bases were uniform) this hash function is 
// uniform. 
//
// 
class MerHashF {
 public:
  enum
    {
      bucket_size = 4,
      min_buckets = 8
    };

  inline size_t operator()(const Mer &M) const {
    uint32 low32 = M.mData & 0xFFFFFFFF;
    uint32 high32 = M.mData >> 32;
    register size_t h = low32 ^ high32;
    //    cout<<M.ToString()<<"\thashf = "<<h<<endl;
    return(h);
  }
	
  bool equals_mer(const Mer &a,const Mer &b){
    //cout<<hex<<"equals_mer: "<<a.mData<<"\t"<<b.mData<<dec<<endl;
    if (a.mData == b.mData) {
      //cout<<"TRUE: "<<true<<endl;
       return(true);
    }else {
      //cout<<"FALSE: "<<true<<endl;
      return(false);
    }
  }

  bool less_mer(const Mer &a,const Mer &b) {
    return(a.mData < b.mData);
  }

  bool operator()(const Mer &a,const Mer &b){
    return(equals_mer(a,b));
    //return(less_mer(a,b));
  }

  Mer canonicalValue(const Mer &M){
    return(M);
  }

};

//------------------------------------------
// * MerSymmetricHashF
//------------------------------------------
// Hash function where Mer and ReverseComplement(Mer)
// have the same hash value. 
// 
// Efficiency note:  
// 
// As written, handling a mer will result in multiple calls to Reverse with same data.  
// For example, a typical application will do:
//
//  it = M2I.find(M2);				
//  if (it != M2I.end()){
//    (*it).second++;
//  }else{
//    M2I[M2] = 1;
//  }
//
//  The find will have to call both the hash function and the equals operator, 
//  both of which will independently compute Reverse of the mer. 
//  The [] operator will similarly have to call both the hash function and the equals operator. 
// 
//  Some efficiency could be gained by providing alternate functions that take M.Reverse as a 
//  parameter.   Is there some way to override operator[] and find() to achieve the desired effect? 
//
//class MerSymmetricHashF : public hash_compare<Mer> {
class MerSymmetricHashF {
 public:
  enum
    {
      bucket_size = 4,
      min_buckets = 8
    };

  inline size_t operator()(const Mer &M) const {		
    register size_t h;
    Mer RM(M);
    RM.Reverse();
    uint64 tempWord;
    // Select the lesser of the two, fwd or reverse complement....
    tempWord = (M.mData <= RM.mData) ? M.mData : RM.mData;
    //    cout<<"MerSymHash: M.mData="<<M.mData<<"\tRM.mData="<<RM.mData<<"\ttempWord="<<tempWord<<endl;
    uint32 low32,high32;
    low32 = tempWord & 0xFFFFFFFF;
    high32 = tempWord >> 32;

    //cout<<"low32="<<low32<<endl;
    //cout<<"high32="<<high32<<endl;
    h = low32 ^ high32;
    //cout<<"h="<<h<<endl;
    return(h);
  }

  // Want equal to be symmetric also...
  bool operator()(const Mer a,const Mer b) const {
    if (a.mData == b.mData) {
      //      cout<<"FWD equal: "<<a.ToString()<<" = "<<b.ToString()<<endl;
      return(true);
    }
    Mer RM(a);
    RM.Reverse();
    if (RM.mData == b.mData) {
      //      cout<<"REV equal: "<<a.ToString()<<" = "<<b.ToString()<<endl;
      return(true);
    }    
    return(false);
  }

  Mer canonicalValue(const Mer &M){
    Mer RM(M);
    RM.Reverse();
    // Select the lesser of the two, fwd or reverse complement....
    if (M.mData <= RM.mData) return(M);
    else return(RM);
  }
};

// 1010 A
// 1011 B
// 1100 C

//------------------------------------------
// * MerWildcardHashF
//------------------------------------------
//  Like MerHashF except that it treats the middle base to be a wildcard. 
//
// 
class MerWildcardHashF {
 public:
  enum
  {
    bucket_size = 4,
    min_buckets = 8
  };

  uint64 wildcardmask;

  //  1  2  3  4  5  6  7  6 5  4  3   2 1
  // 00 00 00 00 00 00 00 00 00 00 00 00 00 
  // 01 23 45 67 89 01 22 10 98 76 54 32 10
  //                   11

  // 29-1 = 28   
  //
  //  1  2      14  15
  // 01 23 .... 67  89
  //            22  22
  //
  MerWildcardHashF(){
    initWildcardMask();
  }

  void initWildcardMask(){
    int center = Mer::mMerSize-1;
    // cout<<"center:"<<center<<endl;
    wildcardmask = 0x03;
    // 0000001000000
    //cout<<"wildcardmask1: "<<wildcardmask<<endl;
    wildcardmask = wildcardmask<<center;
    // 1111110111111
    //cout<<"wildcardmask2: "<<wildcardmask<<endl;
    wildcardmask = ~wildcardmask;
    //    cout<<hex;
    //cout<<"wildcardmask3: "<<wildcardmask<<endl;
  }

  inline size_t operator()(const Mer &M) const {    
    uint64 MaskedData = M.mData & wildcardmask;
    //cout<<hex;
    //cout<<"hash operator:"<<M.mData<<" & "<<wildcardmask<<" = "<<MaskedData<<endl;
    uint32 low32 = MaskedData & 0xFFFFFFFF;
    uint32 high32 = MaskedData >> 32;
    register size_t h = low32 ^ high32;
    //cout<<"h="<<h<<endl;
    return(h);
  }
	
  bool equals_mer(const Mer &a,const Mer &b){
    uint64 MaskedA = a.mData & wildcardmask;
    uint64 MaskedB = b.mData & wildcardmask;
    //cout<<"equals_mer: "<<MaskedA<<"\t"<<MaskedB<<endl;
    if (MaskedA == MaskedB) return(true);
    else return(false);
  }

  /*
  bool less_mer(const Mer &a,const Mer &b) {
    uint64 MaskedA = a.mData & wildcardmask;
    uint64 MaskedB = b.mData & wildcardmask;
    //cout<<"equals_mer: "<<MaskedA<<"\t"<<MaskedB<<endl;
    return(MaskedA < MaskedB);
  }
  */

  bool operator()(const Mer &a,const Mer &b){
    //    return(less_mer(a,b));
    return(equals_mer(a,b));
  }

  Mer canonicalValue(const Mer &M){
    Mer temp;
    temp.mData = M.mData & wildcardmask;
    //cout<<hex;
    //    cout<<"canonical: "<<M.mData<<"\t"<<temp.mData<<endl;
    return(temp);
  }

};

//------------------------------------------
// * MerSymmetricWildcardHashF
//------------------------------------------
//  Like MerHashF except that it treats the middle base to be a wildcard. 
//  Also treat fwd and rev versions the same. 
// 
class MerSymmetricWildcardHashF {
 public:
  enum
  {
    bucket_size = 4,
    min_buckets = 8
  };

  uint64 wildcardmask;
  uint64 rcwildcardmask;

  MerSymmetricWildcardHashF(){
    initWildcardMask();
  }

  void initWildcardMask(){
    int center = Mer::mMerSize-1;
    wildcardmask = 0x03;
    wildcardmask = wildcardmask<<center;
    wildcardmask = ~wildcardmask;
  }

  inline size_t operator()(const Mer &M) const {       
    Mer RM(M);
    RM.Reverse();
    uint64 tempWord;
    // Select the lesser of the two, fwd or reverse complement....
    tempWord = (M.mData <= RM.mData) ? M.mData : RM.mData;

    uint64 MaskedData = tempWord & wildcardmask;
    uint32 low32,high32;
    low32 = MaskedData & 0xFFFFFFFF;
    high32 = MaskedData >> 32;
    register size_t h = low32 ^ high32;

    return(h);
  }

  bool equals_mer(const Mer &a,const Mer &b){        
    uint64 MaskedA = a.mData & wildcardmask;
    uint64 MaskedB = b.mData & wildcardmask;    
    if (MaskedA == MaskedB) return(true);

    Mer RM(a);
    RM.Reverse();    
    uint64 RMMaskedData = RM.mData & wildcardmask;
    if (RMMaskedData == MaskedB) return(true);
    else return(false);
  }

  bool operator()(const Mer &a,const Mer &b){
    return(equals_mer(a,b));
  }

  Mer canonicalValue(const Mer &M){
    Mer RM(M);
    RM.Reverse();
    uint64 tempWord;
    // Select the lesser of the two, fwd or reverse complement....
    tempWord = (M.mData <= RM.mData) ? M.mData : RM.mData;
    uint64 MaskedData = tempWord & wildcardmask;
    Mer temp;
    temp.mData = MaskedData;
    return(temp);
  }
};

#endif 


/* Pieces

  inline size_t operator()(const Mer &M) const {    
    Mer RC(M);
    RC.Reverse();
    uint64 MaskedData = M.mData & wildcardmask;
    uint64 RCMaskedData = RC.mData | ~wildcardmask;
    cout<<hex;
    //cout<<"hash operator:"<<M.mData<<" & "<<wildcardmask<<" = "<<MaskedData<<endl;
    //cout<<"hash operator:"<<RC.mData<<" | ~"<<wildcardmask<<" = "<<RCMaskedData<<endl;
    uint64 tempWord = (MaskedData <= RCMaskedData) ? MaskedData : RCMaskedData;
    uint32 low32 = tempWord & 0xFFFFFFFF;
    uint32 high32 = tempWord >> 32;
    register size_t h = low32 ^ high32;
    cout<<"h="<<h<<endl;
    return(h);
  }

  bool equals_mer(const Mer &a,const Mer &b){    
    uint64 MaskedA = a.mData & wildcardmask;
    uint64 MaskedB = b.mData & wildcardmask;    
    cout<<hex<<"equals_mer: "<<MaskedA<<"\t"<<MaskedB<<endl;
    if (MaskedA == MaskedB) return(true);

    Mer RM(a);
    RM.Reverse();    
    uint64 RMMaskedData = RM.mData | ~wildcardmask;        
    cout<<hex<<"requals_mer: "<<RMMaskedData<<"\t"<<MaskedB<<endl;
    if (RMMaskedData == MaskedB) return(true);
    else return(false);
  }

  bool operator()(const Mer &a,const Mer &b){
    return(equals_mer(a,b));
  }

  Mer canonicalValue(const Mer &M){
    uint64 fwdCanonical = M.mData &wildcardmask;

    Mer RM(M);
    RM.Reverse();    
    uint64 revCanonical = RM.mData | ~wildcardmask;
        

    Mer temp;
    if (fwdCanonical <= revCanonical){
      temp.mData = fwdCanonical;
    }else{
      temp.mData = revCanonical;
    }

    cout<<"canonical temp.mData="<<temp.mData<<endl;

    return(temp);
  }

 */




















